#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <immintrin.h>

double matrix_multiplication_sequential (
    float *a, int a_rows, int a_columns,
    float *b, int b_rows, int b_columns,
    float *c
) {
    for ( int i = 0; i < a_rows; ++i ) {
        for ( int j = 0; j < b_columns; ++j ) {
            
            float result = 0;
            
            for ( int k = 0; k < a_columns; ++k ) {
                result += a[i * a_columns + k] * b[k * b_columns + j];
            }
            
            c[i * a_rows +j] = result;
        }
    }
}

double matrix_multiplication_vectorized (
    float *a, int a_rows, int a_columns,
    float *b, int b_rows, int b_columns,
    float *c
) {
    for ( int i = 0; i < a_rows; ++i ) {
        for ( int j = 0; j < b_columns; ++j ) {
            
            int overflow = a_columns % 8;
            
            
            __m256i indices = _mm256_set_epi32 (
                7 * b_columns,
                6 * b_columns,
                5 * b_columns,
                4 * b_columns,
                3 * b_columns,
                2 * b_columns,
                1 * b_columns,
                0 * b_columns
            );
            
            __m256 result_vector = _mm256_set1_ps ( 0 );
            for ( int k = 0; k < ( a_columns - overflow ); k += 8 ) {
                __m256 a_vector = _mm256_loadu_ps ( &a[i * a_columns + k] );
                // __m256 b_vector = _mm256_set_ps (
                //     b[( k + 7 ) * b_columns + j],
                //     b[( k + 6 ) * b_columns + j],
                //     b[( k + 5 ) * b_columns + j],
                //     b[( k + 4 ) * b_columns + j],
                //     b[( k + 3 ) * b_columns + j],
                //     b[( k + 2 ) * b_columns + j],
                //     b[( k + 1 ) * b_columns + j],
                //     b[( k + 0 ) * b_columns + j]
                // );
                
                __m256 b_vector = _mm256_i32gather_ps (
                    &b[( k + 0) * b_columns + j],
                    indices,
                    1
                );
                
                result_vector = _mm256_add_ps (
                    _mm256_mul_ps ( a_vector, b_vector),
                    result_vector
                );
            }
            
            float tmp[8] = { 0 };
            _mm256_storeu_ps ( tmp, result_vector );
            
            float result = tmp[0] + tmp[1] + tmp[2] + tmp[3] + tmp[4] + tmp[5] + tmp[6] + tmp[7];
            
            for ( int k = a_columns - overflow; k < a_columns; ++k ) {
                result += a[i * a_columns + k] * b[k * b_columns + j];
            }
            
            c[i * a_rows +j] = result;
        }
    }
}

int main ( int argc, char **argv ) {
    int type       = 1;
    int iterations = 10;
    int a_rows     = 8;
    int a_columns  = 16;
    int b_rows     = a_columns;
    int b_columns  = 8;
    
    if ( argc >= 2 ) { type       = atoi ( argv[1] ); }
    if ( argc >= 3 ) { iterations = atoi ( argv[2] ); }
    if ( argc >= 4 ) {
        sscanf ( argv[3], "%dx%dx%d", &a_rows, &a_columns, &b_columns );
        b_rows = a_columns;
    }
    
    float *a = calloc ( a_rows * a_columns, sizeof ( float ) );
    float *b = calloc ( b_rows * b_columns, sizeof ( float ) );
    float *c = calloc ( a_rows * b_columns, sizeof ( float ) );
    
    if ( a == NULL || b == NULL || c == NULL ) {
        return 1;
    }
    
    for ( int i = 0; i < a_rows; ++i ) {
        for ( int j = 0; j < a_columns; ++j ) {
            a[i * a_columns + j] = i + j;
        }
    }
    
    for ( int i = 0; i < b_rows; ++i ) {
        for ( int j = 0; j < b_columns; ++j ) {
            b[i * b_columns + j] = i * j;
        }
    }
    
    clock_t start = clock ( );
    for ( int i = 0; i < iterations; ++i ) {
        switch ( type ) {
            case 0: {
                matrix_multiplication_sequential ( a, a_rows, a_columns, b, b_rows, b_columns, c );
                break;
            }
            case 1: {
                matrix_multiplication_vectorized ( a, a_rows, a_columns, b, b_rows, b_columns, c );
                break;
            }
        }
    }
    clock_t end = clock ( );
    
    double elapsed_in_ms = ( end - start ) * 1. / CLOCKS_PER_SEC * 1000;
    
    printf ( "ELAPSED (ms) = %lf\n", elapsed_in_ms );
    
    free ( a );
    free ( b );
    free ( c );
    
    return 0;
}